@@ -29,6 +29,7 @@ Revision History:
 #include "directions.h"
 #include "BigAlloc.h"
 #include "Util.h"
+#include "AlignerOptions.h"
 
 using namespace std;
 
@@ -39,51 +40,55 @@ using namespace std;
 #endif
 
 ChimericPairedEndAligner::ChimericPairedEndAligner(
-        GenomeIndex         *index_,
-        unsigned            maxReadSize,
-        unsigned            maxHits,
-        unsigned            maxK,
-        unsigned            maxSeedsFromCommandLine,
-        double              seedCoverage,
-		unsigned            minWeightToCheck,
-        bool                forceSpacing_,
-        unsigned            extraSearchDepth,
-        bool                noUkkonen,
-        bool                noOrderedEvaluation,
-		bool				noTruncation,
-        bool                useAffineGap,
-        bool                ignoreAlignmentAdjustmentsForOm,
-		bool				altAwareness,
-        bool                emitALTAlignments_,
-        PairedEndAligner    *underlyingPairedEndAligner_,
-	    unsigned            minReadLength_,
-        int                 maxSecondaryAlignmentsPerContig,
-        int                 maxScoreGapToPreferNonAltAlignment,
-        unsigned            matchReward,
-        unsigned            subPenalty,
-        unsigned            gapOpenPenalty,
-        unsigned            gapExtendPenalty,
-        int                 minScoreRealignment_,
-        int                 minScoreGapRealignmentALT_,
-        int                 minAGScoreImprovement_,
-        BigAllocator        *allocator)
+        GenomeIndex             *index_,
+        unsigned                 maxReadSize,
+        unsigned                 maxHits,
+        unsigned                 maxK,
+        unsigned                 maxSeedsFromCommandLine,
+        double                   seedCoverage,
+		unsigned                 minWeightToCheck,
+        bool                     forceSpacing_,
+        unsigned                 extraSearchDepth_,
+        DisabledOptimizations    disabledOptimizations,  
+        bool                     useAffineGap_,
+        bool                     ignoreAlignmentAdjustmentsForOm,
+		bool				     altAwareness,
+        bool                     emitALTAlignments_,
+        PairedEndAligner        *underlyingPairedEndAligner_,
+	    unsigned                 minReadLength_,
+        int                      maxSecondaryAlignmentsPerContig,
+        int                      maxScoreGapToPreferNonAltAlignment,
+        int                      flattenMAPQAtOrBelow_,
+        bool                     useSoftClipping_,
+        unsigned                 matchReward,
+        unsigned                 subPenalty,
+        unsigned                 gapOpenPenalty,
+        unsigned                 gapExtendPenalty,
+        unsigned                 fivePrimeEndBonus,
+        unsigned                 threePrimeEndBonus,
+        int                      minScoreRealignment_,
+        int                      minScoreGapRealignmentALT_,
+        int                      minAGScoreImprovement_,
+        bool                     enableHammingScoringBaseAligner_,
+        BigAllocator            *allocator)
 		: underlyingPairedEndAligner(underlyingPairedEndAligner_), forceSpacing(forceSpacing_), index(index_), minReadLength(minReadLength_), emitALTAlignments(emitALTAlignments_), 
-           maxKSingleEnd(maxK / 2),
-           minScoreRealignment(minScoreRealignment_), minScoreGapRealignmentALT(minScoreGapRealignmentALT_), minAGScoreImprovement(minAGScoreImprovement_)
+           maxKSingleEnd(maxK / 2), maxKPairedEnd(maxK), extraSearchDepth(extraSearchDepth_),
+           minScoreRealignment(minScoreRealignment_), minScoreGapRealignmentALT(minScoreGapRealignmentALT_), minAGScoreImprovement(minAGScoreImprovement_), useSoftClipping(useSoftClipping_),
+           flattenMAPQAtOrBelow(flattenMAPQAtOrBelow_), enableHammingScoringBaseAligner(enableHammingScoringBaseAligner_), useAffineGap(useAffineGap_)
 {
     // Create single-end aligners.
     singleAligner = new (allocator) BaseAligner(index, maxHits, maxK / 2  /* allocate half to each end instead of letting it float like when they're aligned together */, maxReadSize,
                                                 maxSeedsFromCommandLine, seedCoverage, minWeightToCheck, extraSearchDepth, 
-                                                noUkkonen, noOrderedEvaluation, noTruncation, useAffineGap, 
+                                                disabledOptimizations, useAffineGap, 
                                                 ignoreAlignmentAdjustmentsForOm, altAwareness, emitALTAlignments, maxScoreGapToPreferNonAltAlignment,
                                                 maxSecondaryAlignmentsPerContig, &lv, &reverseLV,
-                                                matchReward, subPenalty, gapOpenPenalty, gapExtendPenalty,
+                                                matchReward, subPenalty, gapOpenPenalty, gapExtendPenalty, fivePrimeEndBonus, threePrimeEndBonus,
                                                 NULL, allocator);
     
     underlyingPairedEndAligner->setLandauVishkin(&lv, &reverseLV);
 
-    ag.init(matchReward, subPenalty, gapOpenPenalty, gapExtendPenalty);
-    reverseAG.init(matchReward, subPenalty, gapOpenPenalty, gapExtendPenalty);
+    ag.init(matchReward, subPenalty, gapOpenPenalty, gapExtendPenalty, fivePrimeEndBonus, threePrimeEndBonus);
+    reverseAG.init(matchReward, subPenalty, gapOpenPenalty, gapExtendPenalty, fivePrimeEndBonus, threePrimeEndBonus);
 
     underlyingPairedEndAligner->setAffineGap(&ag, &reverseAG);
 
@@ -114,7 +119,7 @@ ChimericPairedEndAligner::~ChimericPairedEndAligner()
 }
 
 #ifdef _DEBUG
-extern bool _DumpAlignments;
+extern volatile bool _DumpAlignments;
 #endif // _DEBUG
 
 
@@ -132,9 +137,14 @@ bool ChimericPairedEndAligner::align(
         _int64                *nSingleEndSecondaryResultsForFirstRead,
         _int64                *nSingleEndSecondaryResultsForSecondRead,
         SingleAlignmentResult *singleEndSecondaryResults,     // Single-end secondary alignments for when the paired-end alignment didn't work properly
-        _int64				  maxLVCandidatesForAffineGapBufferSize,
-        _int64				  *nLVCandidatesForAffineGap,
-        PairedAlignmentResult *lvCandidatesForAffineGap
+        _int64                 maxPairedCandidatesForAffineGapBufferSize,
+        _int64                *nPairedCandidatesForAffineGap,
+        PairedAlignmentResult *pairedCandidatesForAffineGap,
+        _int64                 maxSingleCandidatesForAffineGapBufferSize,
+        _int64                *nSingleCandidatesForAffineGapFirstRead,
+        _int64                *nSingleCandidatesForAffineGapSecondRead,
+        SingleAlignmentResult *singleCandidatesForAffineGap,
+        int                    maxK
 	)
 {
 	result->status[0] = result->status[1] = NotFound;
@@ -146,11 +156,14 @@ bool ChimericPairedEndAligner::align(
     result->basesClippedAfter[0] = result->basesClippedAfter[1] = 0;
     result->clippingForReadAdjustment[0] = result->clippingForReadAdjustment[1] = 0;
     result->agScore[0] = result->agScore[1] = 0;
+    result->liftover[0] = result->liftover[1] = false;
     result->agForcedSingleAlignerCall = false;
 
     firstALTResult->status[0] = firstALTResult->status[1] = NotFound;
 
-    *nLVCandidatesForAffineGap = 0;
+    *nPairedCandidatesForAffineGap = 0;
+    *nSingleCandidatesForAffineGapFirstRead = 0;
+    *nSingleCandidatesForAffineGapSecondRead = 0;
 
 	if (read0->getDataLength() < minReadLength && read1->getDataLength() < minReadLength) {
         TRACE("Reads are both too short -- returning");
@@ -177,17 +190,18 @@ bool ChimericPairedEndAligner::align(
         bool fitInSecondaryBuffer = 
 		underlyingPairedEndAligner->align(read0, read1, result, firstALTResult, maxEditDistanceForSecondaryResults, secondaryResultBufferSize, nSecondaryResults, secondaryResults,
             singleSecondaryBufferSize, maxSecondaryAlignmentsToReturn, nSingleEndSecondaryResultsForFirstRead, nSingleEndSecondaryResultsForSecondRead, 
-            singleEndSecondaryResults, maxLVCandidatesForAffineGapBufferSize, nLVCandidatesForAffineGap, lvCandidatesForAffineGap);
+            singleEndSecondaryResults, maxPairedCandidatesForAffineGapBufferSize, nPairedCandidatesForAffineGap, pairedCandidatesForAffineGap, maxSingleCandidatesForAffineGapBufferSize,
+            nSingleCandidatesForAffineGapFirstRead, nSingleCandidatesForAffineGapSecondRead, singleCandidatesForAffineGap, maxKPairedEnd);
 
-        if (*nLVCandidatesForAffineGap > maxLVCandidatesForAffineGapBufferSize) {
+        if (*nPairedCandidatesForAffineGap > maxPairedCandidatesForAffineGapBufferSize) {
             *nSecondaryResults = *nSingleEndSecondaryResultsForFirstRead = *nSingleEndSecondaryResultsForSecondRead = 0;
-            *nLVCandidatesForAffineGap = maxLVCandidatesForAffineGapBufferSize + 1; // So the caller knows it's the paired LV candidate buffer that overflowed
+            *nPairedCandidatesForAffineGap = maxPairedCandidatesForAffineGapBufferSize + 1; // So the caller knows it's the paired LV candidate buffer that overflowed
             return false;
         }
 
         if (!fitInSecondaryBuffer) {
             *nSingleEndSecondaryResultsForFirstRead = *nSingleEndSecondaryResultsForSecondRead = 0;
-            *nLVCandidatesForAffineGap = 0;
+            *nPairedCandidatesForAffineGap = 0;
             *nSecondaryResults = secondaryResultBufferSize + 1; // So the caller knows it's the paired secondary buffer that overflowed
             return false;
         }
@@ -210,12 +224,20 @@ bool ChimericPairedEndAligner::align(
         sumPairScore = result->score[0] + result->score[1];
         sumPairScoreAlt = firstALTResult->score[0] + firstALTResult->score[1];
 
-        // If we have already seen a good ALT pair, don't separate them with by running the single end aligner
+        //
+        // If the command line requests zeroing out low mapqs, do it now.  There's one other place we need to check below.
+        //
+        result->mapq[0] = result->mapq[0] <= flattenMAPQAtOrBelow ? 0 : result->mapq[0];
+        result->mapq[1] = result->mapq[1] <= flattenMAPQAtOrBelow ? 0 : result->mapq[1];
+
+        // If we have already seen a good ALT pair, don't separate them by running the single end aligner
         bool seenBetterAltResult = (firstALTResult->status[0] != NotFound)
                                    && (firstALTResult->status[1] != NotFound)
                                    && (sumPairScoreAlt <= sumPairScore - minScoreGapRealignmentALT);
 
-        if ((result->usedAffineGapScoring[0] || result->usedAffineGapScoring[1]) && maxScore >= minScoreRealignment && !seenBetterAltResult) {
+        bool useAltLiftover = result->liftover[0] && result->liftover[1];
+
+        if ((result->usedAffineGapScoring[0] || result->usedAffineGapScoring[1]) && maxScore >= minScoreRealignment && !seenBetterAltResult && !useAltLiftover) {
             compareWithSingleEndAlignment = true;
         }
         
@@ -229,7 +251,7 @@ bool ChimericPairedEndAligner::align(
 
     int scoreLimitLeft = maxKSingleEnd;
     if (compareWithSingleEndAlignment) {
-        scoreLimitLeft = sumPairScore - 3;
+        scoreLimitLeft = sumPairScore;
         if (result->status[0] != NotFound && result->status[1] != NotFound) {
             result->agForcedSingleAlignerCall = true;   // Only set this if we wouldn't have done it anyway.
         }
@@ -241,6 +263,7 @@ bool ChimericPairedEndAligner::align(
     //
     Read *read[NUM_READS_PER_PAIR] = {read0, read1};
     _int64 *resultCount[2] = {nSingleEndSecondaryResultsForFirstRead, nSingleEndSecondaryResultsForSecondRead};
+    _int64 *affineCandidates[2] = { nSingleCandidatesForAffineGapFirstRead, nSingleCandidatesForAffineGapSecondRead };
 
     SingleAlignmentResult singleResult[NUM_READS_PER_PAIR];
     SingleAlignmentResult firstSingleALTResult[NUM_READS_PER_PAIR];
@@ -248,6 +271,7 @@ bool ChimericPairedEndAligner::align(
     bool chooseSingleEndMapq = true;
     for (int r = 0; r < NUM_READS_PER_PAIR; r++) {
         _int64 singleEndSecondaryResultsThisTime = 0;
+        _int64 singleEndAffineCandidatesThisTime = 0;
 
         if (compareWithSingleEndAlignment) {
             pairAGScore += result->agScore[r];
@@ -270,13 +294,14 @@ bool ChimericPairedEndAligner::align(
             result->clippingForReadAdjustment[r] = 0;
 
             firstALTResult->status[r] = NotFound;   // Don't need to fill in the rest, this suppresses writing it
+            chooseSingleEndMapq = false;
         } else {
             if (compareWithSingleEndAlignment) {
                 // Single-end alignments are not good enough to be considered
                 if (scoreLimitLeft < 0) {
                     break;
                 }
-                int maxKForRead = __min(maxKSingleEnd, __min(result->score[r], scoreLimitLeft));
+                int maxKForRead = __min((int)maxKSingleEnd, __min(result->score[r], scoreLimitLeft));
                 singleAligner->setMaxK(maxKForRead);
             }
 
@@ -284,22 +309,72 @@ bool ChimericPairedEndAligner::align(
             bool fitInSecondaryBuffer =
                 singleAligner->AlignRead(read[r], &singleResult[r], &firstSingleALTResult[r], maxEditDistanceForSecondaryResults,
                     singleSecondaryBufferSize - *nSingleEndSecondaryResultsForFirstRead, &singleEndSecondaryResultsThisTime,
-                    maxSecondaryAlignmentsToReturn, singleEndSecondaryResults + *nSingleEndSecondaryResultsForFirstRead);
+                    maxSecondaryAlignmentsToReturn, singleEndSecondaryResults + *nSingleEndSecondaryResultsForFirstRead,
+                    maxSingleCandidatesForAffineGapBufferSize - *nSingleCandidatesForAffineGapFirstRead, &singleEndAffineCandidatesThisTime,
+                    singleCandidatesForAffineGap + *nSingleCandidatesForAffineGapFirstRead);
+
+            if (singleEndAffineCandidatesThisTime > (maxSingleCandidatesForAffineGapBufferSize - *nSingleCandidatesForAffineGapFirstRead)) {
+                *nSecondaryResults = *nSingleEndSecondaryResultsForFirstRead = *nSingleEndSecondaryResultsForSecondRead = 0;
+                *nSingleCandidatesForAffineGapFirstRead = maxSingleCandidatesForAffineGapBufferSize + 1; // So the caller knows it's the single end candidate buffer that overflowed
+                return false;
+            }
 
             if (!fitInSecondaryBuffer) {
                 *nSecondaryResults = 0;
                 *nSingleEndSecondaryResultsForFirstRead = singleSecondaryBufferSize + 1;
                 *nSingleEndSecondaryResultsForSecondRead = 0;
+                *nSingleCandidatesForAffineGapFirstRead = 0;
                 return false;
             }
 
+            bool usedHammingScoreingBaseAligner = false;
+            if (useSoftClipping && enableHammingScoringBaseAligner) {
+                //
+                // Try Hamming distance scoring to align any unmapped reads
+                //
+                if (singleResult[r].status == NotFound && result->status[r] == NotFound) {
+                    
+                    usedHammingScoreingBaseAligner = true;
+
+                    singleAligner->AlignRead(read[r], &singleResult[r], &firstSingleALTResult[r], maxEditDistanceForSecondaryResults,
+                        singleSecondaryBufferSize - *nSingleEndSecondaryResultsForFirstRead, &singleEndSecondaryResultsThisTime,
+                        maxSecondaryAlignmentsToReturn, singleEndSecondaryResults + *nSingleEndSecondaryResultsForFirstRead,
+                        maxSingleCandidatesForAffineGapBufferSize - *nSingleCandidatesForAffineGapFirstRead, &singleEndAffineCandidatesThisTime,
+                        singleCandidatesForAffineGap + *nSingleCandidatesForAffineGapFirstRead, true);
+
+                    if (singleEndAffineCandidatesThisTime > (maxSingleCandidatesForAffineGapBufferSize - *nSingleCandidatesForAffineGapFirstRead)) {
+                        *nSecondaryResults = *nSingleEndSecondaryResultsForFirstRead = *nSingleEndSecondaryResultsForSecondRead = 0;
+                        *nSingleCandidatesForAffineGapFirstRead = maxSingleCandidatesForAffineGapBufferSize + 1; // So the caller knows it's the single end candidate buffer that overflowed
+                        return false;
+                    }
+
+                    if (!fitInSecondaryBuffer) {
+                        *nSecondaryResults = 0;
+                        *nSingleEndSecondaryResultsForFirstRead = singleSecondaryBufferSize + 1;
+                        *nSingleEndSecondaryResultsForSecondRead = 0;
+                        *nSingleCandidatesForAffineGapFirstRead = 0;
+                        return false;
+                    }
+
+                    _ASSERT(useAffineGap);
+                    singleAligner->alignAffineGap(read[r], &singleResult[r], &firstSingleALTResult[r], 
+                        singleEndAffineCandidatesThisTime, singleCandidatesForAffineGap + *nSingleCandidatesForAffineGapFirstRead);
+                }
+            }
+
             *(resultCount[r]) = singleEndSecondaryResultsThisTime;
+            *(affineCandidates[r]) = singleEndAffineCandidatesThisTime;
 
             if (compareWithSingleEndAlignment) {
-                if (singleResult[r].score != ScoreAboveLimit) {
-                    scoreLimitLeft -= singleResult[r].score;
-                } else {
-                    scoreLimitLeft = ScoreAboveLimit;
+                //
+                // Do not lower scoreLimit for mate if we used Hamming scoring for the base aligner
+                //
+                if (!usedHammingScoreingBaseAligner) {
+                    if (singleResult[r].score != ScoreAboveLimit && singleResult[r].score != BaseAligner::UnusedScoreValue) {
+                        scoreLimitLeft -= singleResult[r].score;
+                    } else {
+                        scoreLimitLeft = ScoreAboveLimit;
+                    }
                 }
                 singleEndAGScore += singleResult[r].agScore;
                 if (result->agScore[r] >= singleResult[r].agScore) {
@@ -315,8 +390,15 @@ bool ChimericPairedEndAligner::align(
             // If the single-end aligner returns a lower MAPQ choose that for the result
             //
             result->mapq[r] = __min(result->mapq[r], singleResult[r].mapq);
-        }
-    }
+
+            //
+            // If the command line requests zeroing out low mapqs, do it now.  There's one other place we need to check below.
+            //  
+            if (result->mapq[r] <= flattenMAPQAtOrBelow) {
+                result->mapq[r] = 0;
+            } // if we flatten
+        } // for each read
+    } // choose single end mapq
 
 
     if (!compareWithSingleEndAlignment || (singleEndAGScore >= pairAGScore + minAGScoreImprovement)) {
@@ -337,6 +419,7 @@ bool ChimericPairedEndAligner::align(
             } else {
                 result->status[r] = singleResult[r].status;
                 result->mapq[r] = singleResult[r].mapq / 3;   // Heavy quality penalty for chimeric reads
+                result->mapq[r] = result->mapq[r] <= 3 ? 0 : result->mapq[r];
                 result->direction[r] = singleResult[r].direction;
                 result->location[r] = singleResult[r].location;
                 result->score[r] = singleResult[r].score;
