@@ -237,7 +237,8 @@ PairedAlignerOptions::PairedAlignerOptions(const char* i_commandLine)
     maxSeedsSingleEnd(DEFAULT_MAX_HITS_FOR_UNDERLYING_SINGLE_END_ALIGNER),
     minScoreRealignment(3),
     minScoreGapRealignmentALT(3),
-    minAGScoreImprovement(15)
+    minAGScoreImprovement(24),
+    enableHammingScoringBaseAligner(true)
 {
 }
 
@@ -269,6 +270,10 @@ void PairedAlignerOptions::usageMessage()
         "       to skip single-end realignment. Default: %d\n"
         "  -eg  min affine gap score improvement needed for single-end alignments to be considered over\n"
         "       paired-end alignments. Default: %d\n"
+        "  -eh  perform Hamming distance scoring to try and map reads that cannot be mapped by both the paired-end\n"
+        "       and the single-end aligner.  This is the default.\n"
+        " -eh-  Do not perform Hamming distance scoring to try and map reads that cannot be mapped by both the paired-end\n"
+        "       and the single-end aligner.\n" 
         ,
         DEFAULT_MIN_SPACING,
         DEFAULT_MAX_SPACING,
@@ -347,9 +352,14 @@ bool PairedAlignerOptions::parse(const char** argv, int argc, int& n, bool *done
             n += 1;
             return true;
         }
+    } else if (strcmp(argv[n], "-eh") == 0) {
+        enableHammingScoringBaseAligner = true;
+        return true;
+    } else if (strcmp(argv[n], "-eh-") == 0) {
+        enableHammingScoringBaseAligner = false;
+        return true;
     }
 
-
     return AlignerOptions::parse(argv, argc, n, done);
 } // PairedAlignerOptions::parse
 
@@ -369,13 +379,15 @@ bool PairedAlignerContext::initialize()
     intersectingAlignerMaxHits = options2->intersectingAlignerMaxHits;
     ignoreMismatchedIDs = options2->ignoreMismatchedIDs;
     quicklyDropUnpairedReads = options2->quicklyDropUnpairedReads;
-    noUkkonen = options->noUkkonen;
-    noOrderedEvaluation = options->noOrderedEvaluation;
+    disabledOptimizations = options->disabledOptimizations;
     inferSpacing = options2->inferSpacing;
     maxSeedsSingleEnd = options2->maxSeedsSingleEnd;
     minScoreRealignment = options2->minScoreRealignment;
     minScoreGapRealignmentALT = options2->minScoreGapRealignmentALT;
-    minAGScoreImprovement = options2->minAGScoreImprovement;
+    useSoftClipping = options->useSoftClipping;
+    minAGScoreImprovement = options->useSoftClipping ? options2->minAGScoreImprovement : 15; // we use an affine gap score threshold of 15 for choosing single end alignments when running with Haplotype Caller (i.e., no soft clipping optimizations)
+    flattenMAPQAtOrBelow = options->flattenMAPQAtOrBelow;
+    enableHammingScoringBaseAligner = options2->enableHammingScoringBaseAligner;
 
 	return true;
 }
@@ -446,6 +458,36 @@ void PairedAlignerContext::computeSpacingDist(GenomeDistance* pairedEndSpacing,
 }
 
 void PairedAlignerContext::runIterationThread()
+{
+    Read* reads[NUM_READS_PER_PAIR];
+    for (int i = 0; i < NUM_READS_PER_PAIR; i++) {
+        reads[i] = NULL;
+    }
+
+#ifdef _MSC_VER
+    __try { // This is the Visual C++ general exception handler.  Alas, it is not standard C++, so we only do this error reporting on Windows.
+#endif // _MSC_VER
+
+        runIterationThreadImpl(reads);
+
+#ifdef _MSC_VER
+    } __except (EXCEPTION_EXECUTE_HANDLER) {
+        if (reads[0] == NULL) {
+            fprintf(stderr, "SNAP crashed while not processing a read.\n");
+        } else {
+            fprintf(stderr, "SNAP crashed processing a read with ID %.*s\n", reads[0]->getIdLength(), reads[0]->getId());
+            for (int whichRead = 0; whichRead < NUM_READS_PER_PAIR; whichRead++) {
+                fprintf(stderr, "@%.*s\n%.*s\n+\n%.*s\n", reads[whichRead]->getIdLength(), reads[whichRead]->getId(), reads[whichRead]->getDataLength(), reads[whichRead]->getData(),
+                        reads[whichRead]->getDataLength(), reads[whichRead]->getQuality());
+            } // for each read
+        } // if we're processing a read
+        fflush(stderr);
+        soft_exit(1);
+    } // except
+#endif // _MSC_VER
+}
+
+void PairedAlignerContext::runIterationThreadImpl(Read **reads)
 {
 	PreventMachineHibernationWhileThisThreadIsAlive();
 
@@ -463,7 +505,6 @@ void PairedAlignerContext::runIterationThread()
 		return;
 	}
 
-    Read *reads[NUM_READS_PER_PAIR];
     _int64 nSingleResults[2] = { 0, 0 };
 
  	if (index == NULL) {
@@ -504,15 +545,16 @@ void PairedAlignerContext::runIterationThread()
 
     int maxReadSize = MAX_READ_LENGTH;
     size_t memoryPoolSize = IntersectingPairedEndAligner::getBigAllocatorReservation(index, intersectingAlignerMaxHits, maxReadSize, index->getSeedLength(), 
-                                                                numSeedsFromCommandLine, seedCoverage, maxDist, extraSearchDepth, maxCandidatePoolSize,
+                                                                numSeedsFromCommandLine, seedCoverage, MAX_K, extraSearchDepth, maxCandidatePoolSize,
                                                                 maxSecondaryAlignmentsPerContig);
 
-    memoryPoolSize += ChimericPairedEndAligner::getBigAllocatorReservation(index, maxReadSize, maxHits, index->getSeedLength(), maxSeedsSingleEnd, seedCoverage, maxDist,
+    memoryPoolSize += ChimericPairedEndAligner::getBigAllocatorReservation(index, maxReadSize, maxHits, index->getSeedLength(), maxSeedsSingleEnd, seedCoverage, MAX_K,
         extraSearchDepth, maxCandidatePoolSize, maxSecondaryAlignmentsPerContig);
 
     _int64 maxPairedSecondaryHits;
     _int64 maxSingleSecondaryHits;
-    _int64 maxPairedLVHitsForAffineGap;
+    _int64 maxPairedCandidatesForAffineGap;
+    _int64 maxSingleCandidatesForAffineGap;
 
     if (maxSecondaryAlignmentAdditionalEditDistance < 0) {
         maxPairedSecondaryHits = 0;
@@ -526,25 +568,28 @@ void PairedAlignerContext::runIterationThread()
     }
 
     if (useAffineGap) {
-        maxPairedLVHitsForAffineGap = 4096;
+        maxPairedCandidatesForAffineGap = 4096;
+        maxSingleCandidatesForAffineGap = 4096;
     }
     else {
-        maxPairedLVHitsForAffineGap = 0;
+        maxPairedCandidatesForAffineGap = 0;
+        maxSingleCandidatesForAffineGap = 0;
     }
 
     bool reallocatedSingleSecondaryBuffer = false;
     bool reallocatedPairedSecondaryBuffer = false;
-    bool reallocatedPairedLVHitsForAffineGapBuffer = false;
+    bool reallocatedPairedCandidatesForAffineGapBuffer = false;
+    bool reallocatedSingleCandidatesForAffineGapBuffer = false;
 
-    memoryPoolSize += (1 + maxPairedSecondaryHits + maxPairedLVHitsForAffineGap) * sizeof(PairedAlignmentResult) + maxSingleSecondaryHits * sizeof(SingleAlignmentResult);
+    memoryPoolSize += (1 + maxPairedSecondaryHits + maxPairedCandidatesForAffineGap) * sizeof(PairedAlignmentResult) + (maxSingleSecondaryHits + maxSingleCandidatesForAffineGap) * sizeof(SingleAlignmentResult);
 
     BigAllocator *allocator = new BigAllocator(memoryPoolSize, 16); // FIXME: Used larger allocation granularity for __m128i that needs to be aligned at 16 byte boundaries
     
-    IntersectingPairedEndAligner *intersectingAligner = new (allocator) IntersectingPairedEndAligner(index, maxReadSize, maxHits, maxDist, numSeedsFromCommandLine, 
+    IntersectingPairedEndAligner *intersectingAligner = new (allocator) IntersectingPairedEndAligner(index, maxReadSize, maxHits, maxDist, maxDistForIndels, numSeedsFromCommandLine, 
                                                                 seedCoverage, minSpacing, maxSpacing, intersectingAlignerMaxHits, extraSearchDepth, 
-                                                                maxCandidatePoolSize, maxSecondaryAlignmentsPerContig, allocator, noUkkonen, noOrderedEvaluation, noTruncation, 
+                                                                maxCandidatePoolSize, maxSecondaryAlignmentsPerContig, allocator, disabledOptimizations, 
                                                                 useAffineGap, ignoreAlignmentAdjustmentForOm, altAwareness, maxScoreGapToPreferNonALTAlignment,
-                                                                matchReward, subPenalty, gapOpenPenalty, gapExtendPenalty);
+                                                                matchReward, subPenalty, gapOpenPenalty, gapExtendPenalty, useSoftClipping);
 
     ChimericPairedEndAligner *aligner = new (allocator) ChimericPairedEndAligner(
         index,
@@ -556,9 +601,7 @@ void PairedAlignerContext::runIterationThread()
 		minWeightToCheck,
         forceSpacing,
         extraSearchDepth,
-        noUkkonen,
-        noOrderedEvaluation,
-		noTruncation,
+        disabledOptimizations,
         useAffineGap,
         ignoreAlignmentAdjustmentForOm,
 		altAwareness,
@@ -567,21 +610,33 @@ void PairedAlignerContext::runIterationThread()
 		minReadLength,
         maxSecondaryAlignmentsPerContig,
         maxScoreGapToPreferNonALTAlignment,
+        flattenMAPQAtOrBelow,
+        useSoftClipping,
         matchReward,
         subPenalty,
         gapOpenPenalty,
         gapExtendPenalty,
+        fivePrimeEndBonus,
+        threePrimeEndBonus,
         minScoreRealignment,
         minScoreGapRealignmentALT,
         minAGScoreImprovement,
+        enableHammingScoringBaseAligner,
         allocator);
 
     allocator->checkCanaries();
 
     PairedAlignmentResult *results = (PairedAlignmentResult *)allocator->allocate((1 + maxPairedSecondaryHits) * sizeof(*results)); // 1 + is for the primary result
-    PairedAlignmentResult *lvCandidatesForAffineGap = (PairedAlignmentResult *)allocator->allocate(maxPairedLVHitsForAffineGap * sizeof(*lvCandidatesForAffineGap));
     SingleAlignmentResult *singleSecondaryResults = (SingleAlignmentResult *)allocator->allocate(maxSingleSecondaryHits * sizeof(*singleSecondaryResults));
 
+    PairedAlignmentResult* pairedCandidatesForAffineGap = NULL;
+    SingleAlignmentResult* singleCandidatesForAffineGap = NULL;
+
+    if (useAffineGap) {
+        pairedCandidatesForAffineGap = (PairedAlignmentResult*)allocator->allocate(maxPairedCandidatesForAffineGap * sizeof(*pairedCandidatesForAffineGap));
+        singleCandidatesForAffineGap = (SingleAlignmentResult*)allocator->allocate(maxSingleCandidatesForAffineGap * sizeof(*singleCandidatesForAffineGap));
+    }
+
     ReadWriter *readWriter = this->readWriter;
 
 #ifdef  _MSC_VER
@@ -600,244 +655,279 @@ void PairedAlignerContext::runIterationThread()
 
     _uint64 readIdxInBatch = 0;
     _int64 startTime = timeInMillis();
-    while (supplier->getNextReadPair(&reads[0],&reads[1])) {
-        _int64 readFinishedTime;
-        if (options->profile) {
-            readFinishedTime = timeInMillis();
-            stats->millisReading += (readFinishedTime - startTime);
-        }
-
-        // Check that the two IDs form a pair; they will usually be foo/1 and foo/2 for some foo.
-        if (!ignoreMismatchedIDs) {
-            Read::checkIdMatch(reads[0], reads[1]);
-        }
-
-        stats->totalReads += 2;
-
-        if (AlignerOptions::useHadoopErrorMessages && stats->totalReads % 10000 == 0 && timeInMillis() - lastReportTime > 10000) {
-            fprintf(stderr, "reporter:counter:SNAP,readsAligned,%llu\n", stats->totalReads - readsWhenLastReported);
-            readsWhenLastReported = stats->totalReads;
-            lastReportTime = timeInMillis();
-        }
-
-        // Skip the pair if there are too many Ns and/or they're too short
-        int maxDist = this->maxDist;
-        bool useful0 = reads[0]->getDataLength() >= minReadLength && (int)reads[0]->countOfNs() <= maxDist;
-        bool useful1 = reads[1]->getDataLength() >= minReadLength && (int)reads[1]->countOfNs() <= maxDist;
-        if (!useful0 && !useful1) {
-            PairedAlignmentResult result;
-            result.status[0] = NotFound;
-            result.status[1] = NotFound;
-            result.location[0] = InvalidGenomeLocation;
-            result.location[1] = InvalidGenomeLocation;
-            nSingleResults[0] = nSingleResults[1] = 0;
-            result.clippingForReadAdjustment[0] = result.clippingForReadAdjustment[1] = 0;
-            result.usedAffineGapScoring[0] = result.usedAffineGapScoring[1] = false;
-            result.basesClippedBefore[0] = result.basesClippedBefore[1] = 0;
-            result.basesClippedAfter[0] = result.basesClippedAfter[1] = 0;
-            result.agScore[0] = result.agScore[1] = 0;
-            result.supplementary[0] = result.supplementary[1] = false;
-
-            bool pass0 = options->passFilter(reads[0], result.status[0], true, false);
-            bool pass1 = options->passFilter(reads[1], result.status[1], true, false);
-            bool pass = (options->filterFlags & AlignerOptions::FilterBothMatesMatch)
-                ? (pass0 && pass1) : (pass0 || pass1);
-
-
-            if (pass) {
-                if (NULL != readWriter) {
-                    readWriter->writePairs(readerContext, reads, &result, 1, NULL, nSingleResults, true, useAffineGap);
-                }
-                stats->uselessReads += 2;
+    while (supplier->getNextReadPair(&reads[0], &reads[1])) {
+            _int64 readFinishedTime;
+            if (options->profile) {
+                readFinishedTime = timeInMillis();
+                stats->millisReading += (readFinishedTime - startTime);
             }
-            else {
+
+            // Check that the two IDs form a pair; they will usually be foo/1 and foo/2 for some foo.
+            if (!ignoreMismatchedIDs) {
+                Read::checkIdMatch(reads[0], reads[1]);
+            }
+
+            stats->totalReads += 2;
+
+            if (AlignerOptions::useHadoopErrorMessages && stats->totalReads % 10000 == 0 && timeInMillis() - lastReportTime > 10000) {
+                fprintf(stderr, "reporter:counter:SNAP,readsAligned,%llu\n", stats->totalReads - readsWhenLastReported);
+                readsWhenLastReported = stats->totalReads;
+                lastReportTime = timeInMillis();
+            }
+
+            // Skip the pair if there are too many Ns and/or they're too short
+            int maxDist = this->maxDist;
+            bool useful0 = reads[0]->getDataLength() >= minReadLength && (int)reads[0]->countOfNs() <= maxDist;
+            bool useful1 = reads[1]->getDataLength() >= minReadLength && (int)reads[1]->countOfNs() <= maxDist;
+            if (!useful0 && !useful1) {
+                PairedAlignmentResult result;
+                result.status[0] = NotFound;
+                result.status[1] = NotFound;
+                result.location[0] = InvalidGenomeLocation;
+                result.location[1] = InvalidGenomeLocation;
+                nSingleResults[0] = nSingleResults[1] = 0;
+                result.clippingForReadAdjustment[0] = result.clippingForReadAdjustment[1] = 0;
+                result.usedAffineGapScoring[0] = result.usedAffineGapScoring[1] = false;
+                result.basesClippedBefore[0] = result.basesClippedBefore[1] = 0;
+                result.basesClippedAfter[0] = result.basesClippedAfter[1] = 0;
+                result.agScore[0] = result.agScore[1] = 0;
+                result.supplementary[0] = result.supplementary[1] = false;
+
+                bool pass0 = options->passFilter(reads[0], result.status[0], true, false);
+                bool pass1 = options->passFilter(reads[1], result.status[1], true, false);
+                bool pass = (options->filterFlags & AlignerOptions::FilterBothMatesMatch)
+                    ? (pass0 && pass1) : (pass0 || pass1);
+
+
+                if (pass) {
+                    if (NULL != readWriter) {
+                        readWriter->writePairs(readerContext, reads, &result, 1, NULL, nSingleResults, true, useAffineGap);
+                    }
+                    stats->uselessReads += 2;
+                }
+                else {
+                    stats->filtered += 2;
+                }
+
+                continue;
+            }
+
+
+            _int64 startTime;
+            if (TIME_HISTOGRAM || options->attachAlignmentTimes) {
+                startTime = timeInNanos();
+            }
+
+            _int64 nSecondaryResults;
+            _int64 nPairedCandidatesForAffineGap;
+            _int64 nSingleSecondaryResults[2];
+            _int64 nSingleCandidatesForAffineGap[2];
+            PairedAlignmentResult firstALTResult;
+
+            while (!aligner->align(reads[0], reads[1], results, &firstALTResult, maxSecondaryAlignmentAdditionalEditDistance, maxPairedSecondaryHits, &nSecondaryResults, results + 1,
+                maxSingleSecondaryHits, maxSecondaryAlignments, &nSingleSecondaryResults[0], &nSingleSecondaryResults[1], singleSecondaryResults,
+                maxPairedCandidatesForAffineGap, &nPairedCandidatesForAffineGap, pairedCandidatesForAffineGap,
+                maxSingleCandidatesForAffineGap, &nSingleCandidatesForAffineGap[0], &nSingleCandidatesForAffineGap[1], singleCandidatesForAffineGap, maxDist)) {
+
+                _ASSERT(nSecondaryResults > maxPairedSecondaryHits || nSingleSecondaryResults[0] > maxSingleSecondaryHits ||
+                    nPairedCandidatesForAffineGap > maxPairedCandidatesForAffineGap || nSingleCandidatesForAffineGap[0] > maxSingleCandidatesForAffineGap);
+
+                if (nSecondaryResults > maxPairedSecondaryHits) {
+                    if (reallocatedPairedSecondaryBuffer) {
+                        BigDealloc(results);
+                        results = NULL;
+                    }
+
+                    maxPairedSecondaryHits *= 2;
+                    results = (PairedAlignmentResult*)BigAlloc((maxPairedSecondaryHits + 1) * sizeof(PairedAlignmentResult));
+                    reallocatedPairedSecondaryBuffer = true;
+                }
+
+                if (nSingleSecondaryResults[0] > maxSingleSecondaryHits) {
+                    if (reallocatedSingleSecondaryBuffer) {
+                        BigDealloc(singleSecondaryResults);
+                        singleSecondaryResults = NULL;
+                    }
+
+                    maxSingleSecondaryHits *= 2;
+                    singleSecondaryResults = (SingleAlignmentResult*)BigAlloc(maxSingleSecondaryHits * sizeof(SingleAlignmentResult));
+                    reallocatedSingleSecondaryBuffer = true;
+                }
+
+                if (nPairedCandidatesForAffineGap > maxPairedCandidatesForAffineGap) {
+                    _ASSERT(useAffineGap);
+                    if (reallocatedPairedCandidatesForAffineGapBuffer) {
+                        BigDealloc(pairedCandidatesForAffineGap);
+                        pairedCandidatesForAffineGap = NULL;
+                    }
+
+                    maxPairedCandidatesForAffineGap *= 2;
+                    pairedCandidatesForAffineGap = (PairedAlignmentResult*)BigAlloc((maxPairedCandidatesForAffineGap) * sizeof(PairedAlignmentResult));
+                    reallocatedPairedCandidatesForAffineGapBuffer = true;
+                }
+
+                if (nSingleCandidatesForAffineGap[0] > maxSingleCandidatesForAffineGap) {
+                    _ASSERT(useAffineGap);
+                    if (reallocatedSingleCandidatesForAffineGapBuffer) {
+                        BigDealloc(singleCandidatesForAffineGap);
+                        singleCandidatesForAffineGap = NULL;
+                    }
+
+                    maxSingleCandidatesForAffineGap *= 2;
+                    singleCandidatesForAffineGap = (SingleAlignmentResult*)BigAlloc(maxSingleCandidatesForAffineGap * sizeof(SingleAlignmentResult));
+                    reallocatedSingleCandidatesForAffineGapBuffer = true;
+                }
+            }
+
+            _int64 alignFinishedTime;
+            if (options->profile) {
+                alignFinishedTime = timeInMillis();
+                stats->millisAligning += (alignFinishedTime - readFinishedTime);
+            }
+
+            _int64 runTime;
+            if (TIME_HISTOGRAM || options->attachAlignmentTimes) {
+                runTime = timeInNanos() - startTime;
+
+                if (options->attachAlignmentTimes) {
+                    if (runTime > 0) {
+                        results[0].alignmentTimeInNanoseconds = runTime;
+                    } else {
+                        results[0].alignmentTimeInNanoseconds = 0;
+                    }
+                } // attachAlignmentTimes
+            } // if we're fine-grained timing reads
+
+#if     TIME_HISTOGRAM
+            if (runTime < 0) { // For reasons that I really don't understand, this seems to run backwards sometimes.  Just ignore the sample when it does.
+                stats->backwardsTimeStamps++;
+                stats->totalBackwardsTimeStamps += runTime;
+            } else {
+                int timeBucket = min(30, cheezyLogBase2(runTime));
+                stats->countByTimeBucket[timeBucket] += 2;
+                stats->nanosByTimeBucket[timeBucket] += runTime;
+
+                for (int whichRead = 0; whichRead < NUM_READS_PER_PAIR; whichRead++) {
+                    if (results[0].status[whichRead] == NotFound) {
+                        stats->countOfUnaligned++;
+                        stats->timeOfUnaligned += runTime / 2;
+                    } else {
+                        stats->countByMAPQ[results[0].mapq[whichRead]]++;
+                        stats->timeByMAPQ[results[0].mapq[whichRead]] += runTime / 2;
+
+                        int score = __min(results[0].score[whichRead], 30);
+                        stats->countByNM[score]++;
+                        stats->timeByNM[score] += runTime / 2;
+                    } // aligned/not aligner
+                } // which read in the pair
+        } // if we got a good time
+#endif // TIME_HISTOGRAM
+
+            if (forceSpacing && isOneLocation(results[0].status[0]) != isOneLocation(results[0].status[1])) {
+                // either both align or neither do
+                results[0].status[0] = results[0].status[1] = NotFound;
+                results[0].location[0] = results[0].location[1] = InvalidGenomeLocation;
+                results[0].usedAffineGapScoring[0] = results[0].usedAffineGapScoring[1] = false;
+                results[0].basesClippedBefore[0] = results[0].basesClippedBefore[1] = 0;
+                results[0].basesClippedAfter[0] = results[0].basesClippedAfter[1] = 0;
+                results[0].agScore[0] = results[0].agScore[1] = 0;
+            }
+
+            bool firstIsPrimary = true;
+            for (int i = 0; i <= nSecondaryResults; i++) {  // Loop runs to <= nSecondaryResults because there's a primary result, too.
+                bool pass0 = options->passFilter(reads[0], results[i].status[0], !useful0, i != 0 || !firstIsPrimary);
+                bool pass1 = options->passFilter(reads[1], results[i].status[1], !useful1, i != 0 || !firstIsPrimary);
+                bool pass = (options->filterFlags & AlignerOptions::FilterBothMatesMatch)
+                    ? (pass0 && pass1) : (pass0 || pass1);
+
+                if (!pass) {
+                    //
+                    // Remove this one from the list by copying the last one here.
+                    //
+                    results[i] = results[nSecondaryResults];
+                    nSecondaryResults--;
+                    if (0 == i) {
+                        firstIsPrimary = false;
+                    }
+                    i--;
+                } else if (options->attachAlignmentTimes) {
+                    results[i].alignmentTimeInNanoseconds = runTime;
+                }
+            }
+
+            //
+            // Now check the single secondary alignments
+            //
+            SingleAlignmentResult* singleResults[2] = { singleSecondaryResults, singleSecondaryResults + nSingleSecondaryResults[0] };
+            for (int whichRead = 0; whichRead < NUM_READS_PER_PAIR; whichRead++) {
+                for (int whichAlignment = 0; whichAlignment < nSingleSecondaryResults[whichRead]; whichAlignment++) {
+                    if (!options->passFilter(reads[whichRead], singleResults[whichRead][whichAlignment].status, false, true)) {
+                        singleResults[whichRead][whichAlignment] = singleResults[whichRead][nSingleSecondaryResults[whichRead] - 1];
+                        nSingleSecondaryResults[whichRead]--;
+                        whichAlignment--;
+                    } else if (options->attachAlignmentTimes) {
+                        singleResults[whichRead]->alignmentTimeInNanoseconds = runTime;
+                    }
+                }
+            }
+
+            if (NULL != readWriter) {
+                readWriter->writePairs(readerContext, reads, results, nSecondaryResults + 1, singleResults, nSingleSecondaryResults, firstIsPrimary, useAffineGap);
+
+                if (emitALTAlignments && (firstALTResult.status[0] != NotFound || firstALTResult.status[1] != NotFound)) {
+                    readWriter->writePairs(readerContext, reads, &firstALTResult, 1, NULL, 0, true, useAffineGap);
+                }
+            }
+
+            if (options->profile) {
+                startTime = timeInMillis();
+                stats->millisWriting += (startTime - alignFinishedTime);
+            }
+
+            stats->extraAlignments += nSecondaryResults + (firstIsPrimary ? 0 : 1); // If first isn't primary, it's secondary.
+
+            if (inferSpacing) {
+                pairedEndSpacing[readIdxInBatch] = 0;
+            }
+
+            if (firstIsPrimary) {
+                updateStats((PairedAlignerStats*)stats, reads[0], reads[1], &results[0], useful0, useful1);
+                if (inferSpacing) {
+                    if ((results[0].direction[0] == FORWARD && results[0].direction[1] == RC) ||
+                        (results[0].direction[0] == RC && results[0].direction[1] == FORWARD)) {
+                        pairedEndSpacing[readIdxInBatch] = DistanceBetweenGenomeLocations(results[0].location[0], results[0].location[1]);
+                        readIdxInBatch++;
+                    }
+                }
+            } else {
                 stats->filtered += 2;
             }
 
-            continue;
-        }
-
-
-#if     TIME_HISTOGRAM
-        _int64 startTime = timeInNanos();
-#endif // TIME_HISTOGRAM
-
-        _int64 nSecondaryResults;
-        _int64 nLVCandidatesForAffineGap;
-        _int64 nSingleSecondaryResults[2];
-        PairedAlignmentResult firstALTResult;
-
-        while (!aligner->align(reads[0], reads[1], results, &firstALTResult, maxSecondaryAlignmentAdditionalEditDistance, maxPairedSecondaryHits, &nSecondaryResults, results + 1,
-            maxSingleSecondaryHits, maxSecondaryAlignments, &nSingleSecondaryResults[0], &nSingleSecondaryResults[1], singleSecondaryResults, maxPairedLVHitsForAffineGap, &nLVCandidatesForAffineGap, lvCandidatesForAffineGap)) {
-
-            _ASSERT(nSecondaryResults > maxPairedSecondaryHits || nSingleSecondaryResults[0] > maxSingleSecondaryHits || nLVCandidatesForAffineGap > maxPairedLVHitsForAffineGap);
-
-            if (nSecondaryResults > maxPairedSecondaryHits) {
-                if (reallocatedPairedSecondaryBuffer) {
-                    BigDealloc(results);
-                    results = NULL;
-                }
-
-                maxPairedSecondaryHits *= 2;
-                results = (PairedAlignmentResult *)BigAlloc((maxPairedSecondaryHits + 1) * sizeof(PairedAlignmentResult));
-                reallocatedPairedSecondaryBuffer = true;
-            }
-
-            if (nSingleSecondaryResults[0] > maxSingleSecondaryHits) {
-                if (reallocatedSingleSecondaryBuffer) {
-                    BigDealloc(singleSecondaryResults);
-                    singleSecondaryResults = NULL;
-                }
-
-                maxSingleSecondaryHits *= 2;
-                singleSecondaryResults = (SingleAlignmentResult *)BigAlloc(maxSingleSecondaryHits * sizeof(SingleAlignmentResult));
-                reallocatedSingleSecondaryBuffer = true;
-            }
-
-            if (nLVCandidatesForAffineGap > maxPairedLVHitsForAffineGap) {
-                if (reallocatedPairedLVHitsForAffineGapBuffer) {
-                    BigDealloc(lvCandidatesForAffineGap);
-                    lvCandidatesForAffineGap = NULL;
-                }
-
-                maxPairedLVHitsForAffineGap *= 2;
-                lvCandidatesForAffineGap = (PairedAlignmentResult *)BigAlloc((maxPairedLVHitsForAffineGap) * sizeof(PairedAlignmentResult));
-                reallocatedPairedLVHitsForAffineGapBuffer = true;
-            }
-        }
-
-        _int64 alignFinishedTime;
-        if (options->profile) {
-            alignFinishedTime = timeInMillis();
-            stats->millisAligning += (alignFinishedTime - readFinishedTime);
-        }
-
-#if     TIME_HISTOGRAM
-        _int64 runTime = timeInNanos() - startTime;
-        if (runTime < 0) { // For reasons that I really don't understand, this seems to run backwards sometimes.  Just ignore the sample when it does.
-            stats->backwardsTimeStamps++;
-            stats->totalBackwardsTimeStamps += runTime;
-        } else {
-            int timeBucket = min(30, cheezyLogBase2(runTime));
-            stats->countByTimeBucket[timeBucket] += 2;
-            stats->nanosByTimeBucket[timeBucket] += runTime;
-
-            for (int whichRead = 0; whichRead < NUM_READS_PER_PAIR; whichRead++) {
-                if (results[0].status[whichRead] == NotFound) {
-                    stats->countOfUnaligned++;
-                    stats->timeOfUnaligned += runTime / 2;
-                }
-                else {
-                    stats->countByMAPQ[results[0].mapq[whichRead]]++;
-                    stats->timeByMAPQ[results[0].mapq[whichRead]] += runTime / 2;
-
-                    int score = __min(results[0].score[whichRead], 30);
-                    stats->countByNM[score]++;
-                    stats->timeByNM[score] += runTime / 2;
-                }
-            }
-        }
-#endif // TIME_HISTOGRAM
-
-        if (forceSpacing && isOneLocation(results[0].status[0]) != isOneLocation(results[0].status[1])) {
-            // either both align or neither do
-            results[0].status[0] = results[0].status[1] = NotFound;
-            results[0].location[0] = results[0].location[1] = InvalidGenomeLocation;
-            results[0].usedAffineGapScoring[0] = results[0].usedAffineGapScoring[1] = false;
-            results[0].basesClippedBefore[0] = results[0].basesClippedBefore[1] = 0;
-            results[0].basesClippedAfter[0] = results[0].basesClippedAfter[1] = 0;
-            results[0].agScore[0] = results[0].agScore[1] = 0;
-        }
-
-        bool firstIsPrimary = true;
-        for (int i = 0; i <= nSecondaryResults; i++) {  // Loop runs to <= nSecondaryResults because there's a primary result, too.
-            bool pass0 = options->passFilter(reads[0], results[i].status[0], !useful0, i != 0 || !firstIsPrimary);
-            bool pass1 = options->passFilter(reads[1], results[i].status[1], !useful1, i != 0 || !firstIsPrimary);
-            bool pass = (options->filterFlags & AlignerOptions::FilterBothMatesMatch)
-                ? (pass0 && pass1) : (pass0 || pass1);
-
-            if (!pass) {
-                //
-                // Remove this one from the list by copying the last one here.
-                //
-                results[i] = results[nSecondaryResults];
-                nSecondaryResults--;
-                if (0 == i) {
-                    firstIsPrimary = false;
-                }
-                i--;
-            }
-        }
-
-        //
-        // Now check the single secondary alignments
-        //
-        SingleAlignmentResult *singleResults[2] = { singleSecondaryResults, singleSecondaryResults + nSingleSecondaryResults[0] };
-        for (int whichRead = 0; whichRead < NUM_READS_PER_PAIR; whichRead++) {
-            for (int whichAlignment = 0; whichAlignment < nSingleSecondaryResults[whichRead]; whichAlignment++) {
-                if (!options->passFilter(reads[whichRead], singleResults[whichRead][whichAlignment].status, false, true)) {
-                    singleResults[whichRead][whichAlignment] = singleResults[whichRead][nSingleSecondaryResults[whichRead] - 1];
-                    nSingleSecondaryResults[whichRead]--;
-                    whichAlignment--;
-                }
-            }
-        }
-
-        if (NULL != readWriter) {
-            readWriter->writePairs(readerContext, reads, results, nSecondaryResults + 1, singleResults, nSingleSecondaryResults, firstIsPrimary, useAffineGap);
-
-            if (emitALTAlignments && (firstALTResult.status[0] != NotFound || firstALTResult.status[1] != NotFound)) {
-                readWriter->writePairs(readerContext, reads, &firstALTResult, 1, NULL, 0, true, useAffineGap);
-            }
-        }
-
-        if (options->profile) {
-            startTime = timeInMillis();
-            stats->millisWriting += (startTime - alignFinishedTime);
-        }
-
-        stats->extraAlignments += nSecondaryResults + (firstIsPrimary ? 0 : 1); // If first isn't primary, it's secondary.
-
-        if (inferSpacing) {
-            pairedEndSpacing[readIdxInBatch] = 0;
-        }
-
-        if (firstIsPrimary) {
-            updateStats((PairedAlignerStats*)stats, reads[0], reads[1], &results[0], useful0, useful1);
             if (inferSpacing) {
-                if ((results[0].direction[0] == FORWARD && results[0].direction[1] == RC) ||
-                    (results[0].direction[0] == RC && results[0].direction[1] == FORWARD)) {
-                    pairedEndSpacing[readIdxInBatch] = DistanceBetweenGenomeLocations(results[0].location[0], results[0].location[1]);
-                    readIdxInBatch++;
+                // Compute new minSpacing and maxSpacing for batch
+                if (readIdxInBatch == DEFAULT_BATCH_SIZE_IS_ESTIMATION) {
+
+                    // Sort all alignments based on spacing
+                    qsort(pairedEndSpacing, DEFAULT_BATCH_SIZE_IS_ESTIMATION, sizeof(GenomeDistance), compareBySpacing);
+
+                    int newMinSpacing = minSpacing, newMaxSpacing = maxSpacing;
+                    double avg, stddev;
+                    computeSpacingDist(pairedEndSpacing, &newMinSpacing, &newMaxSpacing, &avg, &stddev);
+
+                    // fprintf(stderr, "SNAP paired-end read spacing (min, max, avg, stddev) = (%d, %d, %.3f, %.3f)\n", newMinSpacing, newMaxSpacing, avg, stddev);
+
+                    // Update min and max spacing for paired-end aligner
+                    intersectingAligner->setMinSpacing(newMinSpacing);
+                    intersectingAligner->setMaxSpacing(newMaxSpacing);
+
+                    readIdxInBatch = 0;
                 }
             }
-        } else {
-            stats->filtered += 2;
-        }
 
-        if (inferSpacing) {
-            // Compute new minSpacing and maxSpacing for batch
-            if (readIdxInBatch == DEFAULT_BATCH_SIZE_IS_ESTIMATION) {
-
-                // Sort all alignments based on spacing
-                qsort(pairedEndSpacing, DEFAULT_BATCH_SIZE_IS_ESTIMATION, sizeof(GenomeDistance), compareBySpacing);
-
-                int newMinSpacing = minSpacing, newMaxSpacing = maxSpacing;
-                double avg, stddev;
-                computeSpacingDist(pairedEndSpacing, &newMinSpacing, &newMaxSpacing, &avg, &stddev);
-
-                // fprintf(stderr, "SNAP paired-end read spacing (min, max, avg, stddev) = (%d, %d, %.3f, %.3f)\n", newMinSpacing, newMaxSpacing, avg, stddev);
-
-                // Update min and max spacing for paired-end aligner
-                intersectingAligner->setMinSpacing(newMinSpacing);
-                intersectingAligner->setMaxSpacing(newMaxSpacing);
-
-                readIdxInBatch = 0;
-            }
-        }
     }   // while we have a read pair
 
-    stats->lvCalls = aligner->getLocationsScored();
+    stats->lvCalls = aligner->getLocationsScoredWithLandauVishkin();
+    stats->affineGapCalls = aligner->getLocationsScoredWithAffineGap();
 
     allocator->checkCanaries();
 
@@ -851,9 +941,14 @@ void PairedAlignerContext::runIterationThread()
         singleSecondaryResults = NULL;
     }
 
-    if (reallocatedPairedLVHitsForAffineGapBuffer) {
-        BigDealloc(lvCandidatesForAffineGap);
-        lvCandidatesForAffineGap = NULL;
+    if (reallocatedPairedCandidatesForAffineGapBuffer) {
+        BigDealloc(pairedCandidatesForAffineGap);
+        pairedCandidatesForAffineGap = NULL;
+    }
+
+    if (reallocatedSingleCandidatesForAffineGapBuffer) {
+        BigDealloc(singleCandidatesForAffineGap);
+        singleCandidatesForAffineGap = NULL;
     }
 
     aligner->~ChimericPairedEndAligner();
