@@ -1260,7 +1260,10 @@ SortedDataFilterSupplier::mergeSort()
         {
             i->reader = readerSupplier->getDataReader(1, MAX_READ_LENGTH * 8, 0.0,
                 __min(1UL << 23, __max(1UL << 17, bufferSpace / blocks.size()))); // 128kB to 8MB buffer space per block
-            i->reader->init(tempFileName);
+            if (!i->reader->init(tempFileName)) {
+                WriteErrorMessage("SortedDataFilterSupplier::mergeSort: reader->init(%s) failed\n", tempFileName);
+                soft_exit(1);
+            }
             i->reader->reinit(i->start, i->bytes);
         }
     }
@@ -1274,8 +1277,11 @@ SortedDataFilterSupplier::mergeSort()
         DataReader* headerReader;
         if (blocks[0].dataReaderIsBuffer) 
         {
-            headerReader = readerSupplier->getDataReader(1, MAX_READ_LENGTH * 8, 0.0, headerSize + 4096);
-            headerReader->init(tempFileName);
+            headerReader = readerSupplier->getDataReader(1, 0, 0.0, 0);
+            if (!headerReader->init(tempFileName)) {
+                WriteErrorMessage("SortedDataFilterSupplier::mergeSort: reader->init(%s) failed for headerReader\n", tempFileName);
+                soft_exit(1);
+            }
         } else {
             headerReader = blocks[0].reader;
         }
@@ -1289,10 +1295,12 @@ SortedDataFilterSupplier::mergeSort()
 			if ((!headerReader->getData(&rbuffer, &rbytes)) || rbytes == 0) {
                 headerReader->nextBatch();
 				if (!headerReader->getData(&rbuffer, &rbytes)) {
-					WriteErrorMessage( "read header failed\n");
+					WriteErrorMessage( "read header failed, left %lld, headerSize %lld\n", left, headerSize);
+                    headerReader->dumpState();
 					soft_exit(1);
 				}
 			}
+
 			if ((! writer->getBuffer(&wbuffer, &wbytes)) || wbytes == 0) {
 				writer->nextBatch();
 				if (! writer->getBuffer(&wbuffer, &wbytes)) {
@@ -1303,6 +1311,7 @@ SortedDataFilterSupplier::mergeSort()
 			size_t xfer = min(left, min((size_t) rbytes, wbytes));
 			_ASSERT(xfer > 0 && xfer <= UINT32_MAX);
 			memcpy(wbuffer, rbuffer, xfer);
+
             headerReader->advance(xfer);
 			writer->advance((unsigned) xfer);
 			left -= xfer;
